#!/usr/bin/python3

'''
フォントの各情報から、グリフに関する情報をまとめて出力。
$ ./make-ginfo.py <spot-gid> <spot-cmap> [<spot-gsub>] > out.txt

GSUB は省略可。
以下のコマンドで出力したファイルを指定する。

gid: $ spot -nc FONTFILE > gid.txt
cmap: $ spot -t cmap=7 FONTFILE > cmap.txt
gsub: $ spot -t GSUB=7 FONTFILE > gsub.txt

<グリフ名(CID/GID)>TAB<GID>TAB<Unicode(16進数)/...>TAB<このグリフを置換元とする置換先グリフ名(fea:name/...)>
'''

import sys
import re

if len(sys.argv) < 3:
	print('[usage] ' + sys.argv[0] + ' <gid> <cmap> [<gsub>]')
	sys.exit(0)

args = sys.argv[1:]

class ConvData:
	# グリフ名から値を取得
	def _getname(self,name):
		if name[0] == '\\':
			name = int(name[1:])
		elif '@' in name:
			name = int(name.split('@')[0])

		return name
	
	# gid テキスト読み込み
	def read_gid(self,fname):
		f = open(fname, 'rt')
		lines = f.readlines()
		f.close()

		self.namedat = {}
		mo = re.compile(r'\[(.+?)\]\=<(.+?)>')

		for l in lines:
			l = l.strip()
			if (not l) or l[0] != '[': continue

			m = mo.match(l)
			if m:
				gid = int(m.group(1))
				name = self._getname(m.group(2))

				self.namedat[name] = [gid, [], {}]

	# cmap テキスト読み込み
	def read_cmap(self,fname):
		f = open(fname, 'rt')
		lines = f.readlines()
		f.close()

		self.unitoname = {}
		mo = re.compile(r'\[(.+?)\]\=\s*<(.+?)>')

		for l in lines:
			l = l.strip()
			if (not l) or l[0] != '[': continue

			m = mo.match(l)
			if m:
				utxt = m.group(1)
				name = m.group(2)
				# 異体字は除外
				if ' ' in utxt: continue

				uc = int(utxt, base=16)
				name = self._getname(name)

				self.unitoname[uc] = name

				if uc not in self.namedat[name][1]:
					self.namedat[name][1].append(uc)

	# gsub ファイル読み込み
	def read_gsub(self,fname):
		f = open(fname, 'rt')
		lines = f.readlines()
		f.close()

		self.gsubfea = {}
		curname = None
		mo = re.compile(r'sub\s+([^ ]+?)\s+by\s+([^ ]+?);')
		mo_fea = re.compile(r"feature\s+'(.+?)'")

		for l in lines:
			l = l.strip()
			if not l: continue

			if curname:
				if l[0] == '#':
					if l.find('End Lookup') != -1:
						# 空の feature は削除
						if not self.gsubfea[curname]:
							del self.gsubfea[curname]
						curname = None
				else:
					m = mo.match(l)
					if m:
						src = self._getname(m.group(1))
						dst = self._getname(m.group(2))
						if src != dst:
							self.gsubfea[curname][src] = dst
				
			elif l[0] == '#' and l.find('Printing') != -1:
				# feature 開始
				m = mo_fea.search(l)
				if m:
					name = m.group(1)
					if name == 'aalt' or name == 'liga' or name == 'ccmp' or name == 'dlig':
						continue

					curname = name 
					if curname not in self.gsubfea:
						self.gsubfea[curname] = {}

	# GSUB 置換元に、置換先の情報セット
	def set_gsub(self):
		for name,dat in self.namedat.items():
			o = {}
			
			for fea in self.gsubfea:
				# 置換元がある
				if name in self.gsubfea[fea]:
					o[fea] = self.gsubfea[fea][name]

			self.namedat[name][2] = o

	# 出力
	def output(self):
		# name gid unicode gsub('fea:name/...')
		for name,dat in self.namedat.items():
			# unicode
			uclst = ''
			for uc in dat[1]:
				uclst += "{0:X}/".format(uc)
			if uclst:
				uclst = uclst[:-1]

			# GSUB
			gsub = ''
			for fea,dst in dat[2].items():
				gsub += '{0}:{1}/'.format(fea, dst)
			if gsub:
				gsub = gsub[:-1]
		
			print('{0}\t{1}\t{2}\t{3}'.format(name, dat[0], uclst, gsub))


#----

c = ConvData()
c.read_gid(args[0])
c.read_cmap(args[1])

if len(args) >= 3:
	c.read_gsub(args[2])
	c.set_gsub()

c.output()
